__description__ = "a module that houses helper functions for streaming executables"
__author__ = "Reed Essick (reed.essick@ligo.org), Patrick Godwin (patrick.godwin@ligo.org)"
__doc__ = "\n\n".join([__description__, __author__])

#---------------------------------------------------------------------------------------------------

import multiprocessing as mp
import os
import time as TIME

import numpy as np

import traceback

### dependencies not in standard Python libraries
from ligo import segments as ligo_segments

from idq import batch
from idq import calibration
from idq import classifiers
from idq import condor
from idq import configparser
from idq import factories
from idq import features
from idq import io
from idq import logs
from idq import names
from idq import utils
from idq import exceptions

#-------------------------------------------------
### reference variables that may be useful throughout

DEFAULT_STREAM_WORKFLOW = 'fork'
DEFAULT_STREAM_LOG_LEVEL = logs.DEFAULT_LOG_LEVEL
DEFAULT_INITIAL_LOOKBACK = 3600 # one hour

DEFAULT_GPS_END = 2000000000 # set to be +infty but we need something that can be cast to an int

DEFAULT_MIN_NEW_SAMPLES = 1
DEFAULT_MAX_SAMPLES = calibration.DEFAULT_MAX_SAMPLES
DEFAULT_MAX_STRIDE = calibration.DEFAULT_MAX_STRIDE

DEFAULT_PLOT_STRIDE = 10
DEFAULT_MAX_METRIC_SPAN = 60        # one minute
DEFAULT_MAX_LATENCY_SPAN = 4 * 3600 # four hours

#---------------------------------------------------------------------------------------------------

def gps_range(gps_start=None, gps_end=None):
    """
    figure out the GPS range for a streaming analysis
    """
    if gps_start is None:
        gps_start = utils.current_gps_time()
    if gps_end is None:
        gps_end = DEFAULT_GPS_END
    assert gps_end >= gps_start, "gps_end (%.3f) must be at least as large as gps_start (%.3f)!"%(gps_end, gps_start)
    return gps_start, gps_end

def stream(
        config_path,
        gps_start=None,
        gps_end=None,
        verbose=False,
        initial_lookback=DEFAULT_INITIAL_LOOKBACK,
        stream_workflow=DEFAULT_STREAM_WORKFLOW,
        stream_log_level=DEFAULT_STREAM_LOG_LEVEL,
        monitoring_cadence=utils.DEFAULT_MONITORING_CADENCE,
        skip_report=False,
        persist=False,
    ):
    '''
    launch and manage all streaming processes
    '''
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    #---------------------------------------------
    ### set up logging
    #---------------------------------------------
    tag = config.get('general', 'tag')

    rootdir = config.get('general', 'rootdir')
    logdir = names.tag2logdir(tag, rootdir=rootdir)
    traindir = names.tag2traindir(tag, rootdir=rootdir)
    calibratedir = names.tag2calibratedir(tag, rootdir=rootdir)

    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2stream_logname(tag),
        log_level=stream_log_level,
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #---------------------------------------------
    ### check that we don't need to generate intial models, calibration_maps
    #---------------------------------------------

    # make all the reporters we care about for this
    reporter_factory = factories.ReporterFactory()
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'train')))
    ritems = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter = reporter_factory(calibratedir, gps_start, gps_end, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'calibrate')))

    # check that something exists for each classifier
    nicknames = config.get('general', 'classifiers').split()

    if np.any([(trainreporter.preferred(names.nickname2train_topic(nickname)) is None) for nickname in nicknames]):
        logger.info('preferred model not available for at least one classifier. Launching batch training job over [%.3f, %.3f]'%(gps_start-initial_lookback, gps_start))
        try:
            batch.train(gps_start-initial_lookback, gps_start, config_path, verbose=verbose) ### FIXME: this may not block if workflow==condor
        except:
            logger.error('initial training job failed!')
            raise

    if np.any([(calibratereporter.preferred(names.nickname2calibrate_topic(nickname)) is None) for nickname in nicknames]):
        logger.info('preferred CalibrationMap not available for at least one classifier. Launching batch evaluate and calibrate jobs over [%.3f, %.3f]'%(gps_start-initial_lookback, gps_start))
        try:
            ### note, we grab the preferred model within batch.evaluate jobs
            batch.evaluate(gps_start-initial_lookback, gps_start, config_path, verbose=verbose, preferred=True) ### FIXME: this may not block if workflow==condor
        except:
            logger.error('initial evaluate job failed!')
            raise
        try:
            ### note, we do not grab preferred quiver because we expect this to exactly match the previous evalute job
            batch.calibrate(gps_start-initial_lookback, gps_start, config_path, verbose=verbose, preferred=False) ### FIXME: this may not block if workflow==condor
        except:
            logger.error('initial calibrate job failed!')
            raise

    #---------------------------------------------
    ### for processes based on workflow
    #---------------------------------------------

    if stream_workflow=="block":
        raise ValueError('stream cannot manage processes via stream_workflow=block because it must launch several processes simultaneously')

    #---

    elif stream_workflow=='fork':
        logger.info('launching asynchronous tasks in parallel via multiprocessing')

        args = (config_path,) ### set up common arguments
        kwargs = {'gps_start':gps_start, 'gps_end':gps_end, 'verbose':verbose}

        workers = []
        try:
            jobs = [(evaluate, 'evaluate'), (calibrate, 'calibrate'), (timeseries, 'timeseries')]
            if not skip_report:
                jobs.append((report, 'report'))
            for job, name in jobs:
                logger.info('forking %s'%name)
                proc = mp.Process(target=job, name=name, args=args, kwargs=kwargs)
                proc.start()
                workers.append(proc)

            logger.info('forking train')
            kwargs['initial_lookback'] = initial_lookback
            proc = mp.Process(target=train, name='train', args=args, kwargs=kwargs)
            proc.start()
            workers.append(proc)

            if persist: ### stick around and watch forked procs if requested
                num_proc = len(workers)
                manager = utils.CadenceManager(utils.current_gps_time(), stride=monitoring_cadence, delay=0) ### hard coding delay here in case DEFAULT_DELAY is changed...
                while num_proc:
                    logger.info('monitoring %d processes'%num_proc)

                    for i in range(num_proc):
                        proc = workers.pop(0)
                        if proc.is_alive(): ### process is still running
                            workers.append(proc)

                        else:
                            proc.join() ### make sure we send SIGKILL, which this should do
                            logger.info('%s terminated with exitcode=%d'%(proc.name, proc.exitcode))
                            if proc.exitcode:
                                raise RuntimeError('%s terminated with exitcode=%d'%(proc.name, proc.exitcode))

                    num_proc = len(workers)
                    manager.wait() ### wait so we don't spam the OS

        except Exception as e:
            while workers:
                proc = workers.pop()
                proc.terminate()
            raise e

    #---

    elif stream_workflow=='condor':
        logger.info('launching asynchronous tasks in parallel via condor')

        if initial_lookback != 0:
            raise ValueError('cannot currently launch streaming training process with non-zero initial lookback through Condor')

        dag_path = condor.stream_dag(
            gps_start,
            gps_end,
            config_path,
            rootdir,
            skip_report=skip_report,
            **configparser.evalkwargs(**dict(config.items('condor'))) ### extract any other class-adds
        )
        logger.info( 'dag written to '+dag_path )

        exitcode = condor.submit_dag(dag_path, block=True) ### blocks until this is done
        if exitcode:
            raise RuntimeError('non-zero returncode for dag '+dag_path )

        if persist: ### stick around and watch forked procs if requested
            raise NotImplementedError('we currently cannot monitor condor for completion of streaming jobs...')

    #---

    else:
        raise ValueError('stream_workflow=%s not understood'%stream_workflow)

    return None

#-------------------------------------------------

def _train_and_report(conn, classifier, reporter, *args):
    """
    a helper function for forked workflow via multiprocessing
    """
    model = classifier.train(*args)
    reporter.start, reporter.end = model.start, model.end ### update this to match the model's provenance
    path = reporter.report(names.nickname2train_topic(classifier.nickname), model, preferred=True) ### report the final result
    conn.send((path, reporter.start, reporter.end)) ### NOTE: almost identical to batch._train_and_report, but we return more things

def train(config_path, gps_start=None, gps_end=None, verbose=False, initial_lookback=DEFAULT_INITIAL_LOOKBACK):
    '''
    run the training process
    '''
    ### set up gps ranges
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)

    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(config.get('general','rootdir'))
    traindir = names.tag2traindir(tag, rootdir) ### used to store results
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    target_channel = config.get('samples', 'target_channel')
    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))

    dirty_bounds = configparser.config2bounds(config.get('samples', 'dirty_bounds'))
    dirty_window = config.getfloat('samples', 'dirty_window')
    random_rate = config.getfloat('train', 'random_rate')

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2train_logname(tag),
        log_level=config.getint('train','log_level'),
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------

    ### the boundaries for these will be re-set within the main loop, hence, set to zero initially
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'train')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(traindir, gps_start, gps_end)
    segmentreporter = io.SegmentReporter(traindir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actually instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=traindir, **configparser.evalkwargs(**dict(items))))
    klassifiersDict = dict((classifier.nickname,classifier) for classifier in klassifiers)

    ### sort classifiers by how they train (their syntax)
    train_with_quiver = []
    inctrain_with_quiver = []
    train_wout_quiver = [] ### NOTE: implicitly, this means we train with classifier_data and target_times (support for OVL)
    inctrain_wout_quiver = []
    for classifier in klassifiers:
        if classifier.trains_with_quiver:
            if isinstance(classifier, classifiers.IncrementalSupervisedClassifier):
                inctrain_with_quiver.append(classifier)
            else:
                train_with_quiver.append(classifier)
        else:
            if isinstance(classifier, classifiers.IncrementalSupervisedClassifier):
                inctrain_wout_quiver.append(classifier)
            else:
                train_wout_quiver.append(classifier)

    make_quivers = bool(train_with_quiver or inctrain_with_quiver)

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab data
    items = config.items('train data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'train'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)
    time = items.pop('time') ### extract the name of the column we use as time

    sitems = config.items('train stream')
    logger.info('stream_processor -> '+' '.join('%s:%s'%_ for _ in sitems))
    sitems = configparser.evalkwargs(**dict(sitems))

    ### initialize stream object
    stream = StreamProcessor(gps_start, gps_end, classifier_data_factory, logger=logger, buffer_kwargs=items, **sitems)

    # configure segdb if required
    if config.has_option('train data discovery', 'ignore_segdb') and config.getboolean('train data discovery','ignore_segdb'):
        pass ### do not register segdb

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')
        stream.configure_segdb(segdb_url, intersect=segdb_intersect, exclude=segdb_exclude, reporter=segmentreporter)

    # set up common data structures re-used throughout the main loop
    accumulated_target_times = 0
    accumulated_random_times = 0
    umbrella = io.UmbrellaClassifierData(gps_start-initial_lookback, gps_start)  ### start, end will be updated in the main loop
    segs = stream._query_segdb(gps_start-initial_lookback, gps_start)
    umbrella.extend_child(classifier_data_factory(gps_start-initial_lookback, gps_start, segs=segs, **items)) ### make a classifier data and included it

    if train_with_quiver:
        ### get target times and draw from random times
        new_target_times = umbrella.target_times(time, target_channel, target_bounds, segs=segs)
        new_random_times = umbrella.random_times(time, target_channel, dirty_bounds, dirty_window, random_rate, segs=segs)

        ### make a new quiver and add to big quiver
        quiver = factories.QuiverFactory(umbrella, avg_rate_threshold=avg_rate_threshold).labeled(new_target_times, new_random_times)
        quiverreporter = io.QuiverReporter(traindir, 0, 0) ### we'll set the start, end times within the main loop

    ### set up nickanmes for reporting
    train_nicknames = dict((classifier.nickname, names.nickname2train_topic(classifier.nickname)) for classifier in klassifiers)
    for classifier in (inctrain_with_quiver+inctrain_wout_quiver): ### load in the preferred models so incremental stuff picks up where it left off
        classifier.model = trainreporter.retrieve(train_nicknames[nickname], preferred=True)

    #-----------------------------------------
    ### actually run jobs
    #-----------------------------------------
    min_new_samples = config.getint('train', 'min_new_samples') if config.has_option('train', 'min_new_samples') else DEFAULT_MIN_NEW_SAMPLES
    max_samples = config.getint('train', 'max_samples') if config.has_option('train', 'max_samples') else DEFAULT_MAX_SAMPLES
    max_stride = config.getfloat('train', 'max_stride') if config.has_option('train', 'max_stride') else DEFAULT_MAX_STRIDE

    workflow = config.get('train', 'workflow')
    if workflow=='fork':
        if config.has_option('train', 'monitoring_cadence'):
            monitoring_cadence = config.getfloat('train', 'monitoring_cadence')
        else:
            monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
        fork_manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
        workers = [] ### declare this here to avoid a possible race condition when error reporting below
 
    ### start up streaming pipeline
    logger.info( 'starting streaming training' )
    while stream.timestamp < gps_end:

        logger.info('--- train stride: [%.3f, %.3f) ---'%stream.seg)

        ### poll for new data
        ### NOTE: we query for the target channel with target bounds just to make sure there's data present
        ### we route this through stream.poll instead of CD.target_times because the CD objects are created within stream.poll
        ### and would otherwise load ALL TRIGGERS, which can be quite expensive
        new_umbrella, segs = stream.poll(channels=[target_channel], bounds=dirty_bounds)

        timestamp = segs[0][0] if len(segs) else stream.timestamp
        lvtm = utils.livetime(utils.segments_intersection(segs, new_umbrella.segs))
        logger.info( 'acquired %.3f sec of data at %d'%(lvtm, timestamp) )

        ### if there is new samples, create a new umbrella and add to big umbrella
        if lvtm:
            ### get target times and draw from random times
            new_target_times = new_umbrella.target_times(time, target_channel, target_bounds, segs=segs)
            accumulated_target_times += len(new_target_times)
            logger.info( 'identified %d new target times for a total of %d accumulated target times'%(len(new_target_times), accumulated_target_times) )

            new_random_times = new_umbrella.random_times(time, target_channel, dirty_bounds, dirty_window, random_rate, segs=segs)
            accumulated_random_times += len(new_random_times)
            logger.info( 'identified %d new random times for a total of %d accumulated random times'%(len(new_random_times), accumulated_random_times) )

            ### make a new quiver and add to big quiver
            if make_quivers:
                new_quiver = factories.QuiverFactory(new_umbrella, avg_rate_threshold=avg_rate_threshold).labeled(new_target_times, new_random_times)
                if train_with_quiver:
                    quiver.append_quiver(new_quiver)

            ### after generating new times and quivers, restrict segments before adding to the big umbrella
            ### this prevents repeated samples from being added in if/when we add padding to new_umbrella
            new_umbrella.restrict_segments(segs, time=time)
            umbrella.extend_children(new_umbrella.children) ### pull out the children of the new umbrella and add them to the big umbrella

            ### launch a new training job if there is enough data
            if accumulated_target_times+accumulated_random_times < min_new_samples:
                logger.info('%d accumulated times < %d, waiting to launch new training jobs'%(accumulated_target_times+accumulated_random_times, min_new_samples))

            else: ### launch a new training job if there is enough data
                logger.info('%d accumulated times >= %d, launching new training jobs'%(accumulated_target_times+accumulated_random_times, min_new_samples))

                ### NOTE: we never expect accumulated_target_times etc to be too big
                ### so we don't filter them in the same way we filter the other cumulative data structures

                ### no conditional here because we use the big umbrella to define the train start, end below
                logger.info('checking the amount of time contained against max_stride')
                if utils.livetime(umbrella.segs) > max_stride: ### prune umbrella
                    logger.info('removing children from umbrella to reduce the total livetime to %.3f sec'%max_stride)
                    umbrella.strip_child()
                    while utils.livetime(umbrella.segs) > max_stride:
                        umbrella.strip_child()

                logger.info('checking the number of samples contained against max_samples')
                if len(umbrella.target_times(time, target_channel, target_bounds)) > max_samples:
                    logger.info('removing children from umbrella to reduce number of target times to <%d'%max_samples)
                    umbrella.strip_child()
                    while len(umbrella.target_times(time, target_channel, target_bounds)) > max_samples:
                        umbrella.strip_child()

                if train_with_quiver:
                    if utils.livetime(quiver.segs) > max_stride:
                        logger.info('removing early samples from quiver until the total livetime is %.3f sec'%max_stride)
                        ### NOTE: assumes time-ordering
                        quiver.start = quiver[0].gps
                        while utils.livetime(quiver.segs) > max_stride:
                            quiver.popleft()
                            quiver.start = quiver[0].gps

                    if len(quiver) > max_samples: ### prune the quiver
                        logger.info('removing samples from quiver to reduce the total count to %d'%max_samples)
                        ### NOTE: assumes time-ordering of vectors (reasonable) and removes the earliest vectors
                        for _ in range(len(quiver)-max_samples): # we don't just use Quiver(collections.deque)'s max_len option because
                            quiver.popleft()                     # we also want to update the start time of the quiver as needed
                        quiver.start = quiver[0].gps

                train_start = umbrella.start
                train_end = umbrella.end
                logger.info( 'launching training job for [%.3f, %.3f) (new [%.3f, %.3f))'%(train_start, train_end, segs[0][0], segs[-1][1]) )

                ### set timestamp for reporter
                trainreporter.start = train_start
                trainreporter.end = train_end ### this should likewise move along

                if make_quivers: ### record if at least one classifier trains with a quiver
                    quiverreporter.start = train_start
                    quiverreporter.end = train_end
                    if inctrain_with_quiver:
                        path = quiverreporter.report('newquiver', new_quiver)
                        logger.info( 'new_quiver written to '+path )

                    if train_with_quiver:
                        path = quiverreporter.report('quiver', quiver)
                        logger.info( 'quiver written to '+path )

                if workflow == "block": ### train classifiers in series
                    logger.info('training classifiers sequentially')

                    if inctrain_with_quiver:
                        ntarget = accumulated_target_times
                        nrandom = accumulated_random_times
                        for classifier in inctrain_with_quiver:
                            nickname = classifier.nickname
                            logger.info('training %s with new_quiver (%d target_times, %d random_times)'%(nickname, ntarget, nrandom))
                            model = classifier.train(new_quiver)
                            trainreporter.start, trainreporter.end = model.start, model.end ### update reporter to match model provenance
                            logger.info('reporting model')
                            path = trainreporter.report(train_nicknames[nickname], model, preferred=True) ### report the resulting model
                            logger.info( 'model for %s written to %s'%(nickname, path) )

                    if inctrain_wout_quiver:
                        ntarget = len(target_times)
                        lvtm = utils.livetime(new_umbrella.segs) ### FIXME: worry about padding?
                        for classifier in inctrain_wout_quiver:
                            nickname = classifier.nickname
                            logger.info( 'training %s with classifier_data (%d target_times, %.3f sec)'%(nickname, ntarget, lvtm) )
                            model = classifier.train(new_umbrella, target_times)
                            trainreporter.start, trainreporter.end = model.start, model.end ### update reporter to match model provenance
                            logger.info('reporting model')
                            path = trainreporter.report(train_nicknames[nickname], model, preferred=True) ### report the resulting model
                            logger.info( 'model for %s written to %s'%(nickname, path) )

                    if train_with_quiver:
                        ntarget, nrandom = quiver.vectors2num()
                        for classifier in train_with_quiver:
                            nickname = classifier.nickname
                            logger.info( 'training %s with quiver (%d target_times, %d random_times)'%(nickname, ntarget, nrandom) )
                            model = classifier.train(quiver)
                            trainreporter.start, trainreporter.end = model.start, model.end ### update reporter to match model provenance
                            logger.info('reporting model')
                            path = trainreporter.report(train_nicknames[nickname], model, preferred=True) ### report the resulting model
                            logger.info( 'model for %s written to %s'%(nickname, path) )

                    if train_wout_quiver:
                        ### generate this for the whole umbrella instead of just new_umbrella
                        ### FIXME: need to ensure we don't have repeated data in umbrella?
                        all_target_times = umbrella.target_times(time, target_channel, target_bounds)
                        ntarget = len(all_target_times)
                        lvtm = utils.livetime(umbrella.segs) ### FIXME: worry about padding?

                        for classifier in train_wout_quiver:
                            nickname = classifier.nickname
                            logger.info( 'training %s with classifier_data (%d target_times, %.3f sec)'%(nickname, ntarget, lvtm) )
                            model = classifier.train(umbrella, all_target_times)
                            trainreporter.start, trainreporter.end = model.start, model.end ### update reporter to match model provenance
                            logger.info('reporting model')
                            path = trainreporter.report(train_nicknames[nickname], model, preferred=True) ### report the resulting model
                            logger.info( 'model for %s written to %s'%(nickname, path) )

                elif workflow == "fork": ### parallelize via multiprocessing
                    logger.info('training classifiers in parallel via multiprocessing')

                    try:
                        workers = []
                        ### set up all the jobs
                        if inctrain_with_quiver:
                            ntarget = accumulated_target_times
                            nrandom = accumulated_random_times
                            for classifier in inctrain_with_quiver:
                                nickname = classifier.nickname
                                logger.info('training %s with new_quiver (%d target_times, %d random_times)'%(nickname, ntarget, nrandom))
                                conn1, conn2 = mp.Pipe()
                                args = (conn1, classifier, trainreporter, new_quiver)
                                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                                proc.start()
                                conn1.close()
                                workers.append( (proc, conn2, args) )

                        if inctrain_wout_quiver:
                            ntarget = len(target_times)
                            lvtm = utils.livetime(new_umbrella.segs) ### FIXME: worry about padding?
                            for classifier in inctrain_wout_quiver:
                                nickname = classifier.nickname
                                logger.info( 'training %s with classifier_data (%d target_times, %.3f sec)'%(nickname, ntarget, lvtm) )
                                conn1, conn2 = mp.Pipe()
                                args = (conn1, classifier, trainreporter, new_umbrella, target_times)
                                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                                proc.start()
                                conn1.close()
                                workers.append( (proc, conn2, args) )

                        if train_with_quiver:
                            ntarget, nrandom = quiver.vectors2num()
                            for classifier in train_with_quiver:
                                nickname = classifier.nickname
                                logger.info( 'training %s with quiver (%d target_times, %d random_times)'%(nickname, ntarget, nrandom) )
                                conn1, conn2 = mp.Pipe()
                                args = (conn1, classifier, trainreporter, quiver)
                                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                                proc.start()
                                conn1.close()
                                workers.append( (proc, conn2, args) )

                        if train_wout_quiver:
                            ### generate this for the whole umbrella instead of just new_umbrella
                            ### FIXME: need to ensure we don't have repeated data in umbrella?
                            all_target_times = umbrella.target_times(time, target_channel, target_bounds)
                            ntarget = len(all_target_times)
                            lvtm = utils.livetime(umbrella.segs) ### FIXME: worry about padding?
    
                            for classifier in train_wout_quiver:
                                nickname = classifier.nickname
                                logger.info( 'training %s with classifier_data (%d target_times, %.3f sec)'%(nickname, ntarget, lvtm) )
                                conn1, conn2 = mp.Pipe()
                                args = (conn1, classifier, trainreporter, umbrella, all_target_times)
                                proc = mp.Process(target=_train_and_report, args=args, name=nickname)
                                proc.start()
                                conn1.close()
                                workers.append( (proc, conn2, args) )

                        ### monitor forked jobs, cleaning up as they finish
                        while workers:
                            proc, conn2, args = workers.pop()
                            if proc.is_alive(): ### still going
                                workers.append( (proc, conn2, args) ) ### add it back in

                            else: ### process is done
                                if proc.exitcode: ### something went wrong with this process!
                                    raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )
            
                                proc.terminate() ### make sure we clean up the process
                                path, trainreporter.start, trainreporter.end = conn2.recv() ### do this because models may be too big for mp.Connection
                                klassifiersDict[proc.name].model = trainreporter.retrieve(train_nicknames[proc.name])
                                conn2.close()
                                logger.info( 'model for %s written to %s'%(proc.name, path) )

                            fork_manager.wait()

                    except Exception as e:
                        while workers:
                            proc, conn2, _ = workers.pop()
                            proc.terminate()
                            conn2.close()
                        raise e

                elif workflow == "condor": ### parallelize via condor
                    raise NotImplementedError

                else:
                    raise ValueError('workflow=%s not understood!'%workflow)

                ### re-set target_times, random_times to be empty arrays because we've now used them
                accumulated_target_times = 0
                accumulated_random_times = 0

        else:
            logger.info('no identified times, skipping training')

        ### repeat at processing cadence
        stream.flush(retain=0) ### FIXME: worry about padding

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def evaluate(config_path, gps_start=None, gps_end=None, verbose=False):
    '''
    the realtime evaluation routine
    '''
    ### set up gps ranges
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)

    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(config.get('general','rootdir'))
    evaluatedir = names.tag2evaluatedir(tag, rootdir=rootdir)
    traindir = names.tag2traindir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, evaluatedir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    ### get info for building quivers
    target_channel = config.get('samples', 'target_channel')
    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))

    dirty_bounds = configparser.config2bounds(config.get('samples', 'dirty_bounds'))
    dirty_window = config.getfloat('samples', 'dirty_window')

    random_rate = config.getfloat('evaluate', 'random_rate')

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2evaluate_logname(tag),
        log_level=config.getint('evaluate','log_level'),
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------

    ### the boundaries for these will be re-set within the main loop, hence, set to zero initially
    ritems = configparser.evalkwargs(**dict(config.items('evaluate reporting')))
    evaluatereporter = reporter_factory(evaluatedir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'evaluate')))
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'evaluate')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(traindir, gps_start, gps_end)
    segmentreporter = io.SegmentReporter(evaluatedir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actually instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=evaluatedir, **configparser.evalkwargs(**dict(items))))

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab data
    items = config.items('evaluate data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'evaluate'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)
    time = items.pop('time') ### extract the name of the column we use as time

    sitems = config.items('evaluate stream')
    logger.info('stream_processor -> '+' '.join('%s:%s'%_ for _ in sitems))
    sitems = configparser.evalkwargs(**dict(sitems))

    ### initialize stream object
    stream = StreamProcessor(gps_start, gps_end, classifier_data_factory, logger=logger, buffer_kwargs=items, **sitems)

    # configure segdb if required
    if config.has_option('evaluate data discovery', 'ignore_segdb') and config.getboolean('evaluate data discovery','ignore_segdb'):
        pass ### do not register segdb

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')
        stream.configure_segdb(segdb_url, intersect=segdb_intersect, exclude=segdb_exclude, reporter=segmentreporter)

    ### set up nickanmes for reporting
    train_nicknames = dict((classifier.nickname, names.nickname2train_topic(classifier.nickname)) for classifier in klassifiers)
    evaluate_nicknames = dict((classifier.nickname, names.nickname2evaluate_topic(classifier.nickname)) for classifier in klassifiers)

    #-----------------------------------------
    ### actually run jobs
    #-----------------------------------------
    max_samples = config.getint('evaluate', 'max_samples') if config.has_option('evaluate', 'max_samples') else DEFAULT_MAX_SAMPLES

    workflow = config.get('evaluate', 'workflow')
    if workflow=='fork':
        if config.has_option('evaluate', 'monitoring_cadence'):
            monitoring_cadence = config.getfloat('evaluate', 'monitoring_cadence')
        else:
            monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
        fork_manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
        workers = [] ### declare this here to avoid a possible race condition when error reporting below

    ### start up streaming pipeline
    logger.info('starting streaming evaluation')
    while stream.timestamp < gps_end:

        logger.info('--- evaluate stride: [%.3f, %.3f) ---'%stream.seg)

        ### retrieve trained models
        for classifier in klassifiers:
            nickname = classifier.nickname
            model = trainreporter.retrieve(train_nicknames[nickname], preferred=True) # get the next model available
                                                                                      # assumes we poll more frequently than models are produced
            ### update if there is a new model
            if model is not None:
                logger.info( 'updating model for '+nickname )
                classifier.model = model

        ### poll for new data
        ### NOTE: we query for the target channel with target bounds just to make sure there's data present
        ### we route this through stream.poll instead of CD.target_times because the CD objects are created within stream.poll
        ### and would otherwise load ALL TRIGGERS, which can be quite expensive
        new_umbrella, segs = stream.poll(channels=[target_channel], bounds=dirty_bounds)

        timestamp = segs[0][0] if len(segs) else stream.timestamp
        lvtm = utils.livetime(utils.segments_intersection(segs, new_umbrella.segs))
        logger.info( 'acquired %.3f sec of data at %d'%(lvtm, timestamp) )

        ### if there is new samples, create a new umbrella
        if lvtm: ### NOTE: we only condition on whether there is something new to do
                 ### and don't care about min_new_samples, max_samples, max_stride
                 ### we just always evaluate everything, which could cause us to bog down at times...

            evaluate_start = segs[0][0]
            evaluate_end = segs[-1][1]

            ### get target times and draw from random times
            target_times = new_umbrella.target_times(time, target_channel, target_bounds, segs=segs)
            logger.info( 'identified %d target times'%len(target_times) )
            if len(target_times) > max_samples:
                logger.info('retaining %d target times'%max_samples)
                target_times = target_times[-max_samples:]

            random_times = new_umbrella.random_times(time, target_channel, dirty_bounds, dirty_window, random_rate, segs=segs)
            logger.info( 'identified %d random times'%len(random_times) )
            if len(random_times) > max_samples:
                logger.info('retaining %d random times'%max_samples)
                random_times = random_times[-max_samples:]

            ### evaluate classifiers if there are times in which to do evaluation
            if len(target_times) > 0 or len(random_times) > 0:

                ### build a quiver
                quiver = factories.QuiverFactory(new_umbrella, avg_rate_threshold=avg_rate_threshold).labeled(target_times, random_times)

                ### update reporter timestamps
                evaluatereporter.start = evaluate_start
                evaluatereporter.end = evaluate_end
                logger.info( 'launching evaluation job for [%.3f, %.3f)'%(evaluate_start, evaluate_end) )

                if workflow == 'block': ### evaluate with each classifier sequentially
                    logger.info('evaluating classifiers sequentially')

                    for classifier in klassifiers:
                        nickname = classifier.nickname
                        logger.info( 'evaluating quiver with %s'%nickname )
                        quiver = classifier.evaluate(quiver)
                        evaluatereporter.start, evaluatereporter.end = quiver.start, quiver.end ### manage provenance
                        logger.info('reporting quiver')
                        path = evaluatereporter.report(evaluate_nicknames[nickname], quiver, preferred=True)
                        logger.info( 'quiver written to %s'%path )

                elif workflow == 'fork': ### parallelize via multiprocessing
                    logger.info( 'evaluating quiver in parallel via multiprocessing' )

                    try:
                        workers = []
                        ### set up all the jobs
                        for classifier in klassifiers:
                            nickname = classifier.nickname
                            logger.info( 'generating timeseries for %s'%nickname )
                            conn1, conn2 = mp.Pipe()
                            args = (conn1, classifier, evaluatereporter, quiver)
                            proc = mp.Process(target=batch._evaluate_and_report, args=args, name=nickname)
                            proc.start()
                            conn1.close()
                            workers.append( (proc, conn2, args) )

                        ### monitor forked jobs, cleaning up as they finish
                        while workers:
                            proc, conn2, args = workers.pop()
                            if proc.is_alive(): ### still going
                                workers.append( (proc, conn2, args) ) ### add it back in

                            else: ### process is done
                                if proc.exitcode: ### something went wrong with this process!
                                    raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                                proc.terminate() ### make sure we clean up the process
                                logger.info( 'evaluated quiver for %s written to %s'%(proc.name, conn2.recv()) )
                                conn2.close()

                            fork_manager.wait()

                    except Exception as e:
                        while workers:
                            proc, conn2, _ = workers.pop(0)
                            proc.terminate()
                            conn2.close()
                        raise e

                elif workflow == 'condor': ### parallelize via condor
                    raise NotImplementedError

                else:
                    raise ValueError('workflow=%s not understood!'%workflow)

            else:
                logger.info('no identified times, skipping evaluation')

        ### repeat at processing cadence
        stream.flush(retain=0) ### FIXME: worry about padding

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def _calibrate_and_report(conn, classifier, reporter, *args):
    '''
    a helper function for forked workflow via multiprocessing
    '''
    calibrate_kwargs = args[-1]
    args = args[:-1]
    calibmap = classifier.calibrate(*args, **calibrate_kwargs)
    reporter.start, reporter.end = calibmap.start, calibmap.end ### manage provenance
    path = reporter.report(names.nickname2calibrate_topic(classifier.nickname), calibmap, preferred=True) ### report the final result
    conn.send((path, reporter.start, reporter.end)) ### NOTE: almost identical to batch._calibrate_and_report but we return more stuff

def calibrate(config_path, gps_start=None, gps_end=None, verbose=False):
    '''
    calibrate the classifier predictions based on historical data
    '''
    ### set up gps ranges
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)

    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(config.get('general','rootdir'))
    evaluatedir = names.tag2evaluatedir(tag, rootdir=rootdir)
    calibratedir = names.tag2calibratedir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, evaluatedir, calibratedir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2calibrate_logname(tag),
        log_level=config.getint('calibrate','log_level'),
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------

    ### the boundaries for these will be re-set within the main loop, hence, set to zero initially
    ritems = configparser.evalkwargs(**dict(config.items('evaluate reporting')))
    evaluatereporter = reporter_factory(evaluatedir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'calibrate')))
    ritems = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter = reporter_factory(calibratedir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'calibrate')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(calibratedir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actually instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=calibratedir, **configparser.evalkwargs(**dict(items))))
    klassifiersDict = dict((classifier.nickname,classifier) for classifier in klassifiers)

    #-----------------------------------------
    # extract basic parameters for calibration
    #-----------------------------------------
    calibrate_kwargs = configparser.evalkwargs(**dict(config.items('calibrate')))
    min_new_samples = calibrate_kwargs.pop('min_new_samples', DEFAULT_MIN_NEW_SAMPLES)

    # set up cadence manager
    sitems = config.items('evaluate stream') ### NOTE: THIS IS NOT A TYPO. We need the calibration and evaluation jobs to be tightly linked
                                             ### I'm very intentionally taking the stream cadence from the evaluation jobs for the calibration jobs
    logger.info('cadence_manager -> '+' '.join('%s:%s'%_ for _ in sitems))
    sitems = configparser.evalkwargs(**dict(sitems))

    assert 'max_timestamp' not in sitems, 'max_timestamp is set by hand to gps_end and cannot be passed through INI file via evaluate stream'
    manager = utils.CadenceManager(gps_start, max_timestamp=gps_end, logger=logger, **sitems)

    # define nicknames for persistent loop
    evaluate_nicknames = dict((classifier.nickname, names.nickname2evaluate_topic(classifier.nickname)) for classifier in klassifiers)
    calibrate_nicknames = dict((classifier.nickname, names.nickname2calibrate_topic(classifier.nickname)) for classifier in klassifiers)
    for classifier in klassifiers: ### load in existing calibration maps so jobs pick up where they left off
        classifier.calibration_map = calibratereporter.retrieve(calibrate_nicknames[classifier.nickname], preferred=True)

    # set up accumulated quivers as needed
    # NOTE: we only need to do this here because everything else uses StreamProcessor to avoid reading in redundant samples, etc
    quivers = {}
    for classifier in klassifiers:
        quiver = features.Quiver()
        quiver.start = manager.timestamp
        quiver.end = manager.timestamp
        quivers[classifier.nickname] = quiver
    
    #-----------------------------------------
    ### actually run jobs
    #-----------------------------------------
    workflow = config.get('calibrate', 'workflow')
    if workflow=='fork':
        if config.has_option('calibrate', 'monitoring_cadence'):
            monitoring_cadence = config.getfloat('calibrate', 'monitoring_cadence')
        else:
            monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
        fork_manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
        workers = [] ### declare this here to avoid a possible race condition when error reporting below

    ### start up streaming pipeline
    logger.info('starting streaming calibration')
    while manager.timestamp < gps_end:

        logger.info('--- calibrate stride: [%.3f, %.3f) ---'%manager.seg)

        timestamp, _ = manager.wait() ### manage cadence and sleep logic here

        ### read in new data for all classifiers
        for classifier in klassifiers:
            nickname = classifier.nickname
            quiver = quivers[nickname]

            new_quiver = evaluatereporter.retrieve(evaluate_nicknames[nickname], preferred=True) # get the next quiver available
            if new_quiver is None: ### nothing new available
                continue

            elif (len(new_quiver)==0) or (new_quiver.end <= quiver.end):
                logger.info('no new data for %s at %d'%(nickname, timestamp))

            elif new_quiver.start >= quiver.end: ### represents new data!
                ngch, ncln = new_quiver.vectors2num()
                logger.info('acquired %.3f sec of data for %s at %d (%d glitch, %d clean)'%\
                    (utils.livetime(new_quiver.segs), nickname, timestamp, ngch, ncln)
                )
                quiver.append_quiver(new_quiver)

            elif new_quiver.end > quiver.end: ### at least some of the new_quiver is actually new
                trunc = utils.segments_intersection(new_quiver.segs, ligo_segments.segmentlist([(quiver.end, new_quiver.end)]))
                ngch, ncln = new_quiver.vectors2num()
                logger.info('acquired %.3f sec of data for %s at %d (%d glitch, %d clean)'%\
                    (utils.livetime(trunc), nickname, timestamp, ngch, ncln)
                )
                quiver.end = new_quiver.end
                quiver.segs = utils.segments_union(quiver.segs, trunc)
                for vector in new_quiver:
                    if vector.gps > quiver.end: ### only the "new vectors"
                        quiver.append(vector)

        ### launch jobs
        if workflow == 'block':
            logger.info( 'calibrating classifiers with evaluated quivers sequentially' )

            for classifier in klassifiers:
                nickname = classifier.nickname
                quiver = quivers[nickname]

                if len(quiver) >= min_new_samples: ### need to spawn a calibration job for this classifier
                    calibratereporter.start = quiver.start
                    calibratereporter.end = quiver.end

                    ### evaluate quiver and save probabilities to disk
                    ngch, ncln = quiver.vectors2num()
                    logger.info( 'calibrating %s (%d glitch, %d clean)'%(nickname, ngch, ncln) )
                    calibmap = classifier.calibrate(quiver, **calibrate_kwargs) ### this works incrementally be default

                    calibratereporter.start, calibratereporter.end = calibmap.start, calibmap.end ### manage provenance
                    logger.info('reporting calibration')
                    path = calibratereporter.report(calibrate_nicknames[nickname], calibmap, preferred=True)
                    logger.info( 'calibration written to %s'%path )

                    ### re-set quiver
                    start = quiver.end ### take as my starting point where the old quiver left off
                    quiver = features.Quiver()
                    quiver.start = start
                    quiver.end = start
                    quivers[nickname] = quiver

                else:
                    logger.info('not enough new data to re-calibrate %s (%d samples; required at least %d)'%(nickname, len(quiver), min_new_samples))

        elif workflow == 'fork':
            logger.info( 'calibrating in parallel via multiprocessing' )

            try:
                workers = []
                ### set up all the jobs
                for classifier in klassifiers:
                    nickname = classifier.nickname
                    quiver = quivers[nickname]

                    if len(quiver) >= min_new_samples: ### need to spawn a calibration job for this classifier
                        calibratereporter.start = quiver.start
                        calibratereporter.end = quiver.end

                        ### evaluate quiver and save probabilities to disk
                        ngch, ncln = quiver.vectors2num()
                        logger.info( 'calibrating %s (%d glitch, %d clean)'%(nickname, ngch, ncln) )
                        conn1, conn2 = mp.Pipe()
                        args = (conn1, classifier, calibratereporter, quiver, calibrate_kwargs)
                        proc = mp.Process(target=_calibrate_and_report, args=args, name=nickname)
                        proc.start()
                        conn1.close()
                        workers.append( (proc, conn2, args) )

                    else:
                        logger.info('not enough new data to re-calibrate %s (%d samples; required at least %d)'%(nickname, len(quiver), min_new_samples))

                ### monitor forked jobs, cleaning up as they finish
                while workers:
                    proc, conn2, args = workers.pop()
                    if proc.is_alive(): ### still going
                        workers.append( (proc, conn2, args) ) ### add it back in

                    else: ### process is done
                        if proc.exitcode: ### something went wrong with this process!
                            raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                        proc.terminate() ### make sure we clean up the process

                        path, calibratereporter.start, calibratereporter.end = conn2.recv() # do this because CalibMaps may be too big for mp.Connection
                        conn2.close()
                        klassifiersDict[proc.name].calibration_map = calibratereporter.retrieve(calibrate_nicknames[proc.name])
                        logger.info( 'calibration for %s written to %s'%(proc.name, path) )

                        ### re-set quiver
                        quiver = quivers[proc.name]

                        start = quiver.end ### take as my starting point where the old quiver left off
                        quiver = features.Quiver()
                        quiver.start = start
                        quiver.end = start
                        quivers[proc.name] = quiver

                    fork_manager.wait()

            except Exception as e:
                while workers:
                    proc, conn2, _ = workers.pop(0)
                    proc.terminate()
                    conn2.close()
                raise e

        elif workflow == 'condor':
            raise NotImplementedError

        else:
            raise ValueError('workflow=%s not understood!'%workflow)

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def timeseries(config_path, gps_start=None, gps_end=None, verbose=False):
    '''
    the realtime evaluation routine
    '''
    ### set up gps ranges
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)

    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    ### define factories to generate objects
    reporter_factory = factories.ReporterFactory()
    classifier_factory = factories.ClassifierFactory()
    classifier_data_factory = factories.ClassifierDataFactory()

    #-----------------------------------------
    ### extract common parameters
    #-----------------------------------------
    tag = config.get('general', 'tag')

    rootdir = os.path.abspath(config.get('general','rootdir'))
    timeseriesdir = names.tag2timeseriesdir(tag, rootdir=rootdir)
    calibratedir = names.tag2calibratedir(tag, rootdir=rootdir)
    traindir = names.tag2traindir(tag, rootdir=rootdir)
    logdir = names.tag2logdir(tag, rootdir=rootdir)

    for directory in [rootdir, traindir, calibratedir, timeseriesdir, logdir]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    ### extract information necessary for channel names in timeseries files
    assert config.has_option('general', 'instrument'), 'instrument not specified in config file!'
    instrument = config.get('general', 'instrument')

    target_channel = config.get('samples', 'target_channel')
    target_bounds = configparser.config2bounds(config.get('samples', 'target_bounds'))
    frequency = config.get('timeseries data discovery', 'frequency')
    assert frequency in target_bounds, 'must specify a frequency range (called %s) within target_bounds'%frequency
    freq_min, freq_max = target_bounds[frequency]
    assert isinstance(freq_min, int) and isinstance(freq_max, int), 'frequency bounds must be integers!'

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2timeseries_logname(tag),
        log_level=config.getint('timeseries','log_level'),
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #-----------------------------------------
    ### set up how we record results
    #-----------------------------------------

    ### the boundaries for these will be re-set within the main loop, hence, set to zero initially
    ritems = configparser.evalkwargs(**dict(config.items('train reporting')))
    trainreporter = reporter_factory(traindir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))
    ritems = configparser.evalkwargs(**dict(config.items('calibrate reporting')))
    calibratereporter = reporter_factory(calibratedir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))
    ritems = configparser.evalkwargs(**dict(config.items('timeseries reporting')))
    timeseriesreporter = reporter_factory(timeseriesdir, 0, 0, **configparser.add_missing_kwargs(ritems, group=names.tag2group(tag, 'timeseries')))

    ### other reporters used to record intermediate data products
    picklereporter = io.PickleReporter(calibratedir, gps_start, gps_end)
    segmentreporter = io.SegmentReporter(timeseriesdir, gps_start, gps_end)

    #-----------------------------------------
    ### set a random seed if provided
    #-----------------------------------------
    seed = config.getint('samples', 'random_seed') if config.has_option('samples', 'random_seed') else None
    if seed is not None:
        logger.info('setting random seed: %d'%seed )
        np.random.seed(seed)

    #-----------------------------------------
    ### figure out which classifiers we'll run
    #-----------------------------------------
    klassifiers = [] ### funny spelling to avoid conflict with module named "classifiers"
    for nickname in config.get('general', 'classifiers').split():
        items = config.items(nickname)
        logger.info( nickname +' -> ' + ' '.join('%s:%s'%_ for _ in items))

        ### actually instantiate the object and add it to the list
        klassifiers.append(classifier_factory(nickname, rootdir=timeseriesdir, **configparser.evalkwargs(**dict(items))))

    #-----------------------------------------
    ### figure out how we'll read data
    #-----------------------------------------

    ### grab data
    items = config.items('timeseries data discovery')
    logger.info( 'classifier_data -> '+' '.join('%s:%s'%_ for _ in items))
    items = configparser.evalkwargs(**dict(items))
    items = configparser.add_missing_kwargs(items, group=names.tag2group(tag, 'timeseries'))
    avg_rate_threshold = items.pop('avg_rate_threshold', factories.DEFAULT_AVG_RATE_THRESHOLD)

    ### figure out timeseres sampling rate -> dt
    srate = config.getfloat('timeseries', 'srate')
    logger.info('generating timeseries sampled at %.3f Hz'%srate)
    dt = 1./srate ### we pass dt to classifier.timeseries so we only have to do division once.

    ### initialize stream processor
    sitems = config.items('timeseries stream')
    logger.info('stream_processor -> '+' '.join('%s:%s'%_ for _ in sitems))
    sitems = configparser.evalkwargs(**dict(sitems))

    stream = StreamProcessor(gps_start, gps_end, classifier_data_factory, logger=logger, buffer_kwargs=items, **sitems)

    # configure segdb if required
    if config.has_option('timeseries data discovery', 'ignore_segdb') and config.getboolean('timeseries data discovery','ignore_segdb'):
        pass ### do not register segdb

    else:
        segdb_intersect = config.get('segments', 'intersect').split() if config.has_option('segments', 'intersect') else []
        segdb_exclude = config.get('segments', 'exclude').split() if config.has_option('segments', 'exclude') else []
        segdb_url = config.get('segments', 'segdb_url')
        stream.configure_segdb(segdb_url, intersect=segdb_intersect, exclude=segdb_exclude, reporter=segmentreporter)

    train_nicknames = dict((classifier.nickname, names.nickname2train_topic(classifier.nickname)) for classifier in klassifiers)
    calibrate_nicknames = dict((classifier.nickname, names.nickname2calibrate_topic(classifier.nickname)) for classifier in klassifiers)
    timeseries_nicknames = dict((classifier.nickname, names.nickname2timeseries_topic(instrument, classifier.nickname)) for classifier in klassifiers)
    templates = dict((classifier.nickname, names.channel_name_template(instrument, classifier.nickname, freq_min, freq_max)) for classifier in klassifiers)

    #-----------------------------------------
    ### actually run jobs
    #-----------------------------------------
    workflow = config.get('timeseries', 'workflow')
    if workflow=='fork':
        if config.has_option('timeseries', 'monitoring_cadence'):
            monitoring_cadence = config.getfloat('timeseries', 'monitoring_cadence')
        else:
            monitoring_cadence = utils.DEFAULT_MONITORING_CADENCE
        fork_manager = utils.CadenceManager(utils.current_gps_time(), monitoring_cadence, delay=0)
        workers = [] ### declare this here to avoid a possible race condition when error reporting below

    ### start up streaming pipeline
    logger.info('starting streaming timeseries')
    while stream.timestamp < gps_end:

        logger.info('--- timeseries stride: [%.3f, %.3f) ---'%stream.seg)

        # update internals for each classifier
        for classifier in klassifiers:
            nickname = classifier.nickname

            ### retrieve trained model
            model = trainreporter.retrieve(train_nicknames[nickname], preferred=True) # get the next model available
                                                                                      # assumes we poll faster than models are produced
            ### update if there is a new model
            if model is not None:
                logger.info( 'updating model for '+nickname )
                classifier.model = model

            ### retrieve calibration map
            calibration_map = calibratereporter.retrieve(calibrate_nicknames[nickname], preferred=True) # get the next map available
                                                                                                        # assumes we poll faster than maps are produced
            ### update if there is a new calibration map
            if calibration_map:
                logger.info( 'updating calibration map for '+nickname )
                classifier.calibration_map = calibration_map


        ### poll for new data
        ### NOTE: we very intentionally to not "fix_segments" 
        ### keeping them partitioned this way tells us which frames to write 
        ### so that each frame has the same (predictable) duration: stream.stride

        ### NOTE: we query for the target channel with target bounds just to make sure there's data present
        ### we route this through stream.poll instead of CD.target_times because the CD objects are created within stream.poll
        ### and would otherwise load ALL TRIGGERS, which can be quite expensive
        new_umbrella, segs = stream.poll(channels=[target_channel], bounds=target_bounds)

        timestamp = segs[0][0] if len(segs) else stream.timestamp
        lvtm = utils.livetime(utils.segments_intersection(segs, new_umbrella.segs))
        logger.info( 'acquired %.3f sec of data at %d'%(lvtm, timestamp) )

        ### create timeseries if there is new data
        if lvtm:
            ### create umbrella and add to quiver
            quiver_factory = factories.QuiverFactory(new_umbrella, avg_rate_threshold=avg_rate_threshold)

            for seg in segs: ### process a single segment at a time to guarantee uniform duration (stream.stride)
                seglist = ligo_segments.segmentlist([seg])

                ### update report times
                timeseriesreporter.start, timeseriesreporter.end = seg

                if workflow == 'block':
                    logger.info('processing classifiers sequentially')
                    for classifier in klassifiers:
                        nickname = classifier.nickname

                        ### generate series
                        logger.info('generating timeseries for %s'%nickname )
                        series = classifier.timeseries(quiver_factory, dt=dt, segs=seglist)

                        ### calibrate
                        logger.info('calibrating timeseries')
                        seriesdict = calibration.series2dict(series, classifier.model, classifier.calibration_map, templates[nickname])

                        ### report probabilities
                        logger.info('reporting probabilities')
                        for path in batch.report_timeseries(timeseriesreporter, timeseries_nicknames[nickname], seriesdict, preferred=True):
                            logger.info( 'probabilities written to %s'%path )

                        ### record latency
                        time = float(seg[0])
                        latency = utils.gps2latency(time)
                        logger.info( 'latency upon writing probabilities: %.3f'%latency )

                elif workflow == 'fork': ### parallelize via multiprocessing
                    logger.info( 'generating timeseries in parallel via multiprocessing' )

                    try:
                        ### set up all the jobs
                        for classifier in klassifiers:
                            nickname = classifier.nickname
                            logger.info( 'generating timeseries for %s'%nickname )
                            conn1, conn2 = mp.Pipe()
                            args = (conn1, classifier, timeseriesreporter, instrument, (quiver_factory, dt, segs), names.channel_name_template(instrument, nickname, freq_min, freq_max))
                            proc = mp.Process(target=batch._timeseries_and_report, args=args, name=nickname)
                            proc.start()
                            conn1.close()
                            workers.append( (proc, conn2, args) )

                        ### monitor forked jobs, cleaning up as they finish
                        while workers:
                            proc, conn2, args = workers.pop()
                            if proc.is_alive(): ### still going
                                workers.append( (proc, conn2, args) ) ### add it back in

                            else: ### process is done
                                if proc.exitcode: ### something went wrong with this process!
                                    raise RuntimeError('non-zero returncode for %s (pid=%d)'%(proc.name, proc.pid) )

                                proc.terminate() ### make sure we clean up the process
                                for path in conn2.recv():
                                    logger.info( 'timeseries for %s written to %s'%(proc.name, path) )
                                conn2.close()

                                ### record latency
                                time = float(seg[0])
                                latency = utils.gps2latency(time)
                                logger.info( 'latency upon writing probabilities: %.3f'%latency )

                            fork_manager.wait()

                    except Exception as e:
                        while workers:
                            proc, conn2, _ = workers.pop(0)
                            proc.terminate()
                            conn2.close()
                        raise e

                elif workflow == 'condor': ### parallelize via condor
                    raise NotImplementedError

                else:
                    raise ValueError('workflow=%s not understood!'%workflow)

        else:
            logger.info('no identified times, skipping timeseries')

        ### repeat at processing cadence
        stream.flush(retain=0) ### FIXME: worry about padding

    #--- not strictly necessary, but it's nice to end with a return statement...
    return None

#-------------------------------------------------

def report(config_path, gps_start=None, gps_end=None, reportdir=None, verbose=False):
    """run batch.report jobs periodically
    """
    ### set up gps ranges
    gps_start, gps_end = gps_range(gps_start=gps_start, gps_end=gps_end)

    ### read in config
    config_path = os.path.abspath(os.path.realpath(config_path))
    config = configparser.path2config(config_path)

    #-----------------------------------------
    ### set up logging
    #-----------------------------------------
    tag = config.get('general', 'tag')
    rootdir = os.path.abspath(config.get('general','rootdir'))

    if reportdir is None:
        logdir = names.tag2logdir(tag, rootdir=rootdir)
    else:
        logdir = reportdir ### write log into the output directory

    if verbose:
        print("writing log : %s"%logdir)

    logger = logs.get_logger(
        names.tag2report_logname(tag),
        log_level=config.getint('report','log_level'),
        rootdir=logdir,
        verbose=verbose,
        rotating=logs.DEFAULT_ROTATION_CADENCE,
    )
    logger.info( "using config : %s"%config_path )

    #---------------------------------------------
    ### set up cadence manager
    #---------------------------------------------
    # support a running average if specified
    lookback = config.getfloat('report', 'lookback') if config.has_option('report', 'lookback') else 0

    sitems = config.items('report stream') ### NOTE: THIS IS NOT A TYPO. We need the calibration and evaluation jobs to be tightly linked
                                             ### I'm very intentionally taking the stream cadence from the evaluation jobs for the calibration jobs
    logger.info('cadence_manager -> '+' '.join('%s:%s'%_ for _ in sitems))
    sitems = configparser.evalkwargs(**dict(sitems))

    assert 'max_timestamp' not in sitems, 'max_timestamp is set by hand to gps_end and cannot be passed through INI file via evaluate stream'
    manager = utils.CadenceManager(gps_start, max_timestamp=gps_end, logger=logger, **sitems)

    #---------------------------------------------
    ### actually loop
    #---------------------------------------------

    logger.info('starting streaming reporting')
    while manager.timestamp < gps_end:

        logger.info('--- report stride: [%.3f, %.3f) ---'%manager.seg)
        start, end = manager.wait() ### manage cadence and sleep logic here

        try:
            ### we allow users to generate a running-average with non-trivial overlap via "lookback"
            batch.report(start-lookback, end, config_path, reportdir=reportdir, verbose=verbose)

        except KeyboardInterrupt:
            raise ### we want to raise this one

        except: # but catch anything else
            logger.warn('report job failed!')
            logger.warn(traceback.format_exc().strip("\n"))

#-------------------------------------------------

class StreamProcessor(object):
    """
    handles data processing logic for the streaming pipeline
    """
    # define a few string templates for things that will be printed repeated
    _poll_template = 'querying for triggers within [%.4f, %.4f)'
    _retained_template = 'retained %.4f sec of livetime'
    _ignore_template = 'ignoring segdb'
    _yesdata_template = 'found data at %.4f'
    _nodata_template = 'no data found within [%.4f, %.4f)'
    _incontiguous_template = 'gap in data found [%.4f, %.4f). Moving to %.4f'
    _latency_template = 'too far behind realtime at (%.4f sec at %.4f), skipping ahead to %.4f'
    _newbuffer_template = 'creating new buffer for [%.4f, %.4f)'

    def __init__(self, start, end, buffer_factory, buffer_kwargs={}, logger=None, **manager_kwargs):

        ### record things
        self._logger = logger # hold on to this reference so we can report about what happens internally here...
                              # this is kinda janky, but it will work...
        self._init_segdb()
        self._buffer_factory = buffer_factory

        ### set up cadence manager
        assert (not 'max_timestamp' in manager_kwargs), 'max_timestamp is set to end by hand, so it cannot be passed via manager_kwargs!'
        self._manager = utils.CadenceManager(start, max_timestamp=end, logger=logger, **manager_kwargs)

        ### do this next little backflip to make sure all options are passed correctly
        self._buffer_kwargs = buffer_kwargs
        self._buffer_kwargs.update(self._buffer_factory(start, end, **buffer_kwargs).kwargs)
        self._buffer = None

        ### set up the main umbrella we manage within our big loop
        self._umbrella = io.UmbrellaClassifierData(self.timestamp, self.timestamp, **self._buffer_kwargs)

    def _init_segdb(self):
        """
        set up all the attributes needed for segdb stuff
        these will all be overwritten upon a call to configure_segdb
        """
        self._ignore_segdb = True ### default behavior
        self._segdb_url = None
        self._segdb_intersect = None
        self._segdb_exclude = None
        self._segdb_reporter = None

    @property
    def ignore_segdb(self):
        return self._ignore_segdb

    @property
    def stride(self):
        return self._manager.stride
    @property
    def delay(self):
        return self._manager.delay

    @property
    def timestamp(self):
        return self._manager.timestamp

    @property
    def seg(self):
        return self._manager.seg

    @property
    def segs(self):
        return self._manager.segs

    @property
    def umbrella(self):
        return self._umbrella

    @property
    def segdb_url(self):
        return self._segdb_url

    @property
    def segdb_intersect(self):
        return self._segdb_intersect

    @property
    def segdb_exclude(self):
        return self._segdb_exclude

    @property
    def segdb_reporter(self):
        return self._segdb_reporter

    def configure_segdb(self, segdb_url, intersect=[], exclude=[], reporter=None):
        """
        we have to configure_segdb in order to not ignore_segdb, and we assume that one would only configure_segdb if you wanted to use it
        thus, we manage the boolean flag controlling whether to use segdb based on whether this method was called
        """
        self._ignore_segdb = False
        self._segdb_url = segdb_url
        self._segdb_intersect = intersect
        self._segdb_exclude = exclude
        self._segdb_reporter = reporter

        self._segdb_template = "querying %s within %s for: intersect=%s ; exclude=%s"%(segdb_url, '[%.3f, %.3f)', ','.join(intersect), ','.join(exclude))

        ### update the existing buffer
        if self._buffer is not None:
            self._buffer.restrict_segments(self._query_segdb(self._buffer.start, self._buffer._end))

    def _query_segdb(self, start, end):
        """
        a helper function that manages segment construction for each buffer
        """
        if self.ignore_segdb:
            if self._logger is not None:
                self._logger.info(self._ignore_template)
            segs = ligo_segments.segmentlist([(start, end)])

        else:
            if self._logger is not None:
                self._logger.info(self._segdb_template%(start, end))

            segs = utils.segdb2segs(
              start,
              end,
              intersect=self._segdb_intersect,
              exclude=self._segdb_exclude,
              segdb_url=self._segdb_url,
              logger=self._logger,
           )

            if self._logger is not None:
                self._logger.info(self._retained_template%utils.livetime(segs))

            if self._segdb_reporter is not None:
                self._segdb_reporter.start = start
                self._segdb_reporter.end = end
                path = self._segdb_reporter.report('segments', segs)
                self._logger.info( 'segments written to '+path )

        return segs

    def new_buffer(self, start):
        end = start+self.stride
        if self._logger is not None:
            self._logger.info(self._newbuffer_template%(start, end))
        return self._buffer_factory(start, end, segs=self._query_segdb(start, end), **self._buffer_kwargs)

    def flush(self, retain=0):
        ### clear out all but the most recently added child (in case new samples need to access older data to create feature vectors)
        while len(self._umbrella.children) > retain:
            self._umbrella.strip_child()

    def wait(self):
        """
        delegates to CadenceManager.wait(), which updates timestamp after sleeping
        """
        return self._manager.wait()

    def poll(self, **triggers_kwargs):
        """
        the workhorse method
        manages requests for data and I/O via ClassifierData objects
        constructs and returns an umbrella of the requested ClassifierDatas for this stride

        triggers_kwargs is passed to calls to ClassifierData.triggers and is exposed to help control I/O costs
        """
        log = self._logger is not None
        segs = ligo_segments.segmentlist([])
        try:
            for seg in self._manager.poll(): ### manager.poll will gobble up multiple segments if needed
                self._buffer = self.new_buffer(seg[0]) ### we wait to create this until after self.wait() because queries to segdb 
                                                  ### within new_buffer() must be causal
                if utils.livetime(self._buffer.segs): ### there is non-trivial livetime in the new buffer
                    if log:
                        self._logger.info(self._poll_template%(self._buffer.start, self._buffer.end))

                    try: ### try to retrieve data
                        data = self._buffer.triggers(**triggers_kwargs)

                    except exceptions.IncontiguousDataError as e: ### span of data doesn't match span of classifier data
                       if log:
                            self._logger.warn(self._incontiguous_template%(e.gap_start, e.gap_end, e.timestamp))
                       self._manager.timestamp = e.timestamp

                    except exceptions.NoDataError as e: ### no data found when querying
                        if log:
                            self._logger.warn(self._nodata_template%(e.timestamp, e.timestamp+e.stride))

                    else:
                        if log:
                            self._logger.info(self._yesdata_template%seg[0])
                        self._umbrella.extend_child(self._buffer)

                segs.append(seg)

        except exceptions.LatencyError as e:
            if self._logger is not None:
                self._logger.warn(self._latency_template%(e.latency, e.current_timestamp, e.timestamp))
            self._manager.timestamp = e.timestamp

        ### return children of umbrella corresponding to new times
        return self._umbrella, segs ### FIXME: worry about padding and edge effects for FeatureVecotr.vectorize?
